/*
c32load: An S19-file loader for the M68MOD912C32 microcontroller

Copyright (C) 2005 Paul Marks

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <fcntl.h>
#include <termios.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/ioctl.h>

#define SDEVICE "/dev/ttyUSB0"  // Change this to match your serial port
#define DISABLE_DTR 1           // If 1 doesn't work, try 0.


#define VERSION "1.0.4"
#define BAUDRATE B115200   // the serial monitor expects 115,200
#define _POSIX_SOURCE 1

#define BUFLEN 512
#define NO_ERROR 0xE0

#define MSG_OK  " \033[0;92mok\033[0m\n"      // 92=green
#define MSG_ERR " \033[0;91merror!\033[0m\n"  // 91=red

int doComm(FILE *, int);
int loadS19(FILE *, int);
int s19_error(int);
int writeBlock(int, int, unsigned const char *, int);
int getStatus(int);
int serialInit(struct termios *);

int main(int argc, const char *argv[])
{
    FILE *sfile;
    int fd, ret;
    struct termios oldtio;

    if (argc != 2)
    {
        printf(
            "c32load-%s: An S19-file loader for the M68MOD912C32 "
            "microcontroller\n\n", VERSION);
        printf(
            "Copyright (C) 2005 Paul Marks\n"
            "http://www.pmarks.net/\n\n"
            "This is free software; see the source for copying conditions.  "
            "There is NO\n"
            "warranty; not even for MERCHANTABILITY or FITNESS FOR A "
            "PARTICULAR PURPOSE.\n\n");

        printf("Usage: %s s19_file\n", argv[0]);
        return -1;
    }

    sfile = fopen(argv[1], "r");
    if (sfile == NULL)
    {
        printf("Can't open file: %s\n", argv[1]);
        return -1;
    }

    printf("Opening %s...", SDEVICE); fflush(stdout);
    fd = serialInit(&oldtio);
    if (fd < 0)  // if the port couldn't be opened
    {
        printf(MSG_ERR);
        perror(SDEVICE);
        fclose(sfile);
        return -1;
    }
    printf(MSG_OK);

    ret = doComm(sfile, fd);

    // cleanup
    fclose(sfile);
    tcsetattr(fd, TCSANOW, &oldtio);
    close(fd);

    return ret;
}

int doComm(FILE *sfile, int fd)
{
    char c;

    printf("Handshaking..."); fflush(stdout);
    c = 0x0D;
    if (write(fd,&c,1) != 1 || getStatus(fd) != NO_ERROR)
    {
        printf(MSG_ERR "-Check connections, set switch to BOOT, and reset.\n");
        return -1;
    }
    printf(MSG_OK);

    printf("Erasing..."); fflush(stdout);
    c = 0xB6;
    if (write(fd,&c,1) != 1 || getStatus(fd) != NO_ERROR)
    {
        printf(MSG_ERR);
        return -1;
    }
    printf(MSG_OK);

    printf("Sending Program..."); fflush(stdout);
    if (loadS19(sfile, fd))
    {
        return -1;
    }
    printf(MSG_OK);

    return 0;
}

int loadS19(FILE *sfile, int fd)
{
    char line[BUFLEN];
    char *lptr;

    unsigned char dataBuf[256];
    int lnum, i;

    char snum;
    int slen, addr, byte;
    int remap_len;

    // read the S19 file, line by line
    for(lnum=1; fgets(line,BUFLEN,sfile); lnum++)
    {
        // parse the first 8 bytes of the line
        if (sscanf(line, "S%c%02X%04X", &snum, &slen, &addr) != 3)
            return s19_error(lnum);

        switch (snum)
        {
            case '0':
                continue; // comment line, go to next
            case '1':
                break;          // normal data, keep going
            case '9':
                return 0;   // EOF marker, stop
            default:
                return s19_error(lnum);
        }

        slen -= 3;  // subtract 2 for address, 1 for checksum

        if (strlen(line) < slen*2+10) // don't want to read any garbage
            return s19_error(lnum);

        // read all the data bytes into an array
        for (i=0, lptr=line+8; i<slen; i++, lptr+=2)
        {
            if (sscanf(lptr, "%02X", &byte) != 1)
                return s19_error(lnum);
            dataBuf[i] = byte;
        }

        // Due to some sort of bug in the serial monitor, I need to remap
        // 8000-BFFF into 4000-7FFF. Thanks to Karl Lunt for providing me
        // with this info.
        remap_len = 0;
        if (addr >= 0x8000 && addr <= 0xBFFF)
        {
            // Look out for a block crossing the C000 boundary
            if (addr+slen > 0xC000)
            {
                // Separate the piece after the boundary
                remap_len = addr+slen-0xC000;
                slen -= remap_len;
            }
            // Remap the rest
            addr -= 0x4000;
        }

        // Write the normal block
        if (writeBlock(fd, addr, dataBuf, slen))
            return -1;

        // This only happens if a block crosses the C000 boundary
        if (remap_len)
            if (writeBlock(fd, 0xC000, dataBuf+slen, remap_len))
                return -1;
    }
    return 0;
}

int s19_error(int line)
{
    printf(MSG_ERR "-Bad S19 data on line %d.\n", line);
    return -1;
}

int writeBlock(int fd, int addr, unsigned const char *data, int len)
{
    unsigned char buf[BUFLEN];
    int wCount;

    // build up the command sequence
    buf[0] = 0xA8;                // $A8 indicates a block write
    buf[1] = addr>>8;                // address, high byte
    buf[2] = addr;                // address, low byte
    buf[3] = len-1;                // writing 1..256 bytes
    memcpy(buf+4, data, len);        // fill in the data

    wCount = write(fd, buf, len+4);

    if (wCount != len+4 || getStatus(fd) != NO_ERROR)
    {
        printf(MSG_ERR "-Problem writing to flash at $%04X\n", addr);
        return -1;
    }
    return 0;
}

int getStatus(int fd)
{
    int i;
    unsigned char status[3];
    for (i = 0; i < 3; i++) {
        if (read(fd, status + i, 1) != 1) {
            return -1;
        }
    }
    return status[0];
}

int serialInit(struct termios *oldtio)
{
    int fd;
    int dtr = TIOCM_DTR;
    struct termios newtio;

    fd = open(SDEVICE, O_RDWR | O_NOCTTY);
    if (fd < 0) return fd;

    tcgetattr(fd, oldtio);  // save the old settings

    memset(&newtio, 0, sizeof(newtio));
    newtio.c_cflag = BAUDRATE | CS8 | CLOCAL | CREAD;
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;
    newtio.c_lflag = 0;

    newtio.c_cc[VTIME] = 20;  // 2-second timeout
    newtio.c_cc[VMIN] = 0;

    tcflush(fd, TCIFLUSH);  // clear the line
    tcsetattr(fd, TCSANOW, &newtio); // activate settings

    // clear the DTR pin.  Some MCUs don't work when it's on.
    if (DISABLE_DTR) {
        ioctl(fd, TIOCMBIC, &dtr);
    }

    return fd;
}
